import sys
import numpy as np

import tensorflow as tf
import tensorflow_addons as tfa

#Custom libraries - User defined
from ..utils import conv_utils
from ..layers import poolSpectral
from ..utils.data_utils import parseDataset
from ..utils.conv_utils import convSpec2d_block, conv2d_block
from ..utils.conv_utils import cross_correlate_iff, cross_correlate_ff, cross_correlate

def unet(input_channel, image_size, filter_size, n_depth, dp_rate = 0.1, act = 'relu'):
    
    input_size = (image_size, image_size, input_channel)
    inputsA = tf.keras.Input(shape=input_size)   # input for CBED
    inputsB = tf.keras.Input(shape=input_size)   # input for probe
    skips = []

    print("Cross correlation layer ... \n")
    cc1 = tf.keras.layers.Lambda(cross_correlate)([inputsA, inputsB])
    #cc1 = tf.keras.layers.BatchNormalization()(cc1)
    
    print("\n")
    print("Building the U-net ... \n")
    
    pool = conv2d_block(cc1, filter_size, n_depth=n_depth, dp_rate=dp_rate, activation=act) 
    
    # conv u-net for 4DSTEM cbed pattern
    for encode in range(int(np.log2(image_size))-1):
        conv = conv2d_block(pool, np.minimum(filter_size * (2**encode), 256), n_depth=n_depth, dp_rate=dp_rate, activation=act)
        pool = tf.keras.layers.MaxPool2D(pool_size=(2, 2))(conv)
        skips.append(conv)
    
    for decode in reversed(range(int(np.log2(256))-1)):
        conv = conv2d_block(pool, np.minimum(filter_size * (2**(decode+1)), 256), n_depth=n_depth, dp_rate=dp_rate, activation=act)
        pool = tf.keras.layers.UpSampling2D(size=(2, 2))(conv)
        pool = tf.keras.layers.Concatenate(axis=-1)([skips[decode], pool])

    unet_final = conv2d_block(pool, filter_size, n_depth=n_depth, dp_rate=dp_rate, activation=act) 
    
    print("\n")
    print("Building the one more convolution layers ... \n")
    conv = conv2d_block(unet_final, filter_size, n_depth=n_depth, dp_rate=dp_rate, activation=act)

    print("\n")
    print("Building the final layer ... \n")
    pred_out = tf.keras.layers.Conv2D(filters = 1, kernel_size = (3, 3),padding='same', kernel_initializer = 'he_uniform')(conv)
    pred_out = tf.keras.layers.Activation('linear', dtype='float32')(pred_out)

    model = tf.keras.Model(inputs=[inputsA, inputsB], outputs=pred_out)

    return model


import functools

import numpy as np

from tensorflow import int32
from tensorflow import cast
from tensorflow import concat
from tensorflow import constant
from tensorflow import less_equal
from tensorflow import transpose
from tensorflow import complex64
from tensorflow import expand_dims
from tensorflow import math
from tensorflow import dtypes
from tensorflow.signal import fft2d
from tensorflow.signal import rfft2d
from tensorflow.signal import ifft2d
from tensorflow.signal import irfft2d
from tensorflow.signal import fftshift
from tensorflow.signal import ifftshift
from tensorflow.random import uniform
from tensorflow.linalg import normalize

from tensorflow.python.framework import tensor_shape
from tensorflow.python.keras import backend
from tensorflow.python.keras.engine.base_layer import Layer
from tensorflow.python.keras.engine.input_spec import InputSpec
from tensorflow.python.keras.utils import conv_utils
from tensorflow.python.ops import array_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import nn
from tensorflow.python.util.tf_export import keras_export

"""
This class (SpectralPool2D) does frequency pooling by converting image into fourier space - works with spectral conv2d (ConvSpectral2D)

Input images should be in real space - this layer does fourier transform of images before the pooling

"""

class SpectralMaxPool2D(Layer):
  """ Spectral pooling layer for arbitrary pooling functions, for 2D inputs (e.g. images).
  Modified the spectral pool function using tf.fft2d replacing tf.ifft2d in release 0.0.2
  """

  def __init__(self, img_size, 
               alpha = 0.3, 
               beta = 0.15 , 
               gamma = 1.0, 
               kappa = 0.5,
               data_format='channels_last',
               dropout = False,
               name=None, **kwargs):
    super(SpectralMaxPool2D, self).__init__(name=name, **kwargs)
    if data_format is None:
      data_format = backend.image_data_format()
    self.data_format = conv_utils.normalize_data_format(data_format)
    self.input_spec = InputSpec(ndim=4)
    self.alpha = alpha
    self.beta = beta
    self.gamma = gamma
    self.kappa = kappa
    self.img_size = img_size
    self.dropout = dropout
 
  def call(self, inputs):
    inputs_fft = transpose(inputs, [0,3,1,2])
    im_fft = fft2d(cast(inputs_fft, complex64))
    filter_size = self._get_sp_dim(self.img_size)
    #filter_size = math.multiply(self.gamma, filter_size)
    im_transformed = self._spectral_pool(im_fft, filter_size)
    
    if self.dropout:
        freq_dropout_lower_bound, freq_dropout_upper_bound = self._get_frq_dropout_bounds(filter_size, self.kappa)
        if (freq_dropout_lower_bound is not None and
            freq_dropout_upper_bound is not None):
            tf_random_cutoff = uniform([], freq_dropout_lower_bound, freq_dropout_upper_bound)
            dropout_mask = self._frequency_dropout_mask(filter_size,
                                                        tf_random_cutoff)
        
            im_downsampled = im_transformed * dropout_mask
        else:
            im_downsampled = im_transformed
    else:
        im_downsampled = im_transformed
    
    outputs = transpose(math.real(ifft2d(im_downsampled)), [0,2,3,1])
    
    return outputs
    
    
  def _spectral_pool(self, images, filter_size):
    assert len(images.get_shape().as_list()) == 4
    assert filter_size >= 3

    if filter_size % 2 == 1:
        n = int((filter_size-1) // 2)
        top_left = images[:, :, :n+1, :n+1]
        top_right = images[:, :, :n+1, -n:]
        bottom_left = images[:, :, -n:, :n+1]
        bottom_right = images[:, :, -n:, -n:]
        top_combined = concat([top_left, top_right], axis=-1)
        bottom_combined = concat([bottom_left, bottom_right], axis=-1)
        all_together = concat([top_combined, bottom_combined], axis=-2)
    else:
        n = int(filter_size // 2)
        top_left = images[:, :, :n, :n]
        top_middle = expand_dims(
            cast(0.5 ** 1.0, complex64) *
            (images[:, :, :n, n] + images[:, :, :n, -n]),
            -1)
        top_right = images[:, :, :n, -(n-1):]
        middle_left = expand_dims(
            cast(0.5 ** 1.0, complex64) *
            (images[:, :, n, :n] + images[:, :, -n, :n]), 
            -2)
        middle_middle = expand_dims(
            expand_dims(
                cast(0.5, complex64) *
                (images[:, :, n, n] + images[:, :, n, -n] +
                images[:, :, -n, n] + images[:, :, -n, -n]),
                -1), -1)
        middle_right = expand_dims(
            cast(0.5 ** 1.0, complex64) *
            (images[:, :, n, -(n-1):] + images[:, :, -n, -(n-1):]),
            -2)
        bottom_left = images[:, :, -(n-1):, :n]
        bottom_middle = expand_dims(
            cast(0.5 ** 1.0, complex64) *
            (images[:, :, -(n-1):, n] + images[:, :, -(n-1):, -n]),
            -1)
        bottom_right = images[:, :, -(n-1):, -(n-1):]
        top_combined = concat(
            [top_left, top_middle, top_right],
            axis=-1)
        middle_combined = concat(
            [middle_left, middle_middle, middle_right],
            axis=-1)
        bottom_combined = concat(
            [bottom_left, bottom_middle, bottom_right],
            axis=-1)
        all_together = concat(
            [top_combined, middle_combined, bottom_combined],
            axis=-2)
    return all_together
    
  def _get_frq_dropout_bounds(self, n, m):
    c = self.alpha + m* (self.beta - self.alpha)
    freq_dropout_lower_bound = c * (1. + n // 2)
    freq_dropout_upper_bound = (1. + n // 2)

    return freq_dropout_lower_bound, freq_dropout_upper_bound

  def _get_sp_dim(self, n):
        if n % 2 == 1:
            fsize = int((n-1) // 2)
        else:
            fsize = int(n // 2)
        # minimum size is 3:
        return max(3, fsize)

  def _frequency_dropout_mask(self, height, frequency_to_truncate_above):
    cutoff_shape = frequency_to_truncate_above.get_shape().as_list()
    assert len(cutoff_shape) == 0

    mid = int(height // 2)
    if height % 2 == 1:
        go_to = mid + 1
    else:
        go_to = mid
    indexes = np.concatenate((
        np.arange(go_to),
        np.arange(mid, 0, -1)
    )).astype(np.float32)

    xs = np.broadcast_to(indexes, (height, height))
    ys = np.broadcast_to(np.expand_dims(indexes, -1), (height, height))
    highest_frequency = np.maximum(xs, ys)
    print(frequency_to_truncate_above)

    comparison_mask = constant(highest_frequency)
    dropout_mask = cast(less_equal(
        comparison_mask,
        frequency_to_truncate_above
    ), complex64)
    
    return dropout_mask

  def compute_output_shape(self, input_shape):
    input_shape = tensor_shape.TensorShape(input_shape).as_list()
    if self.data_format == 'channels_first':
      rows = input_shape[2]
      cols = input_shape[3]
    else:
      rows = input_shape[1]
      cols = input_shape[2]
    rows = conv_utils.conv_output_length(rows, self.pool_size[0], self.padding,
                                         self.strides[0])
    cols = conv_utils.conv_output_length(cols, self.pool_size[1], self.padding,
                                         self.strides[1])
    if self.data_format == 'channels_first':
      return tensor_shape.TensorShape(
          [input_shape[0], input_shape[1], rows, cols])
    else:
      return tensor_shape.TensorShape(
          [input_shape[0], rows, cols, input_shape[3]])

  def get_config(self):
    config = {
        'img_size': self.img_size,
        'alpha': self.alpha,
        'beta': self.beta,
        'gamma': self.gamma,
        'kappa': self.kappa,
        'data_format': self.data_format,
        'name':self.name
    }
    base_config = super(SpectralMaxPool2D, self).get_config()
    return dict(list(base_config.items()) + list(config.items()))



"""
This class (Complex SpectralPool2D) does frequency (Spectral pooling) pooling of the images already in fourier space - works with ConvComplex2D

The input to this pooling layer is a complex fourier image (augmented real and imaginary values such as dimension of batch of images - [batch_size, H, W, input_channel*2] where the input_channel*2 are for real and imag part of the fft of image)

"""

class ComplexSpectralMaxPool2D(Layer):
  """ Spectral pooling layer for arbitrary pooling functions, for 2D complex inputs (e.g. fourier images).
  Modified the spectral pool function using tf.fft2d replacing tf.ifft2d in release 0.0.2
  """

  def __init__(self, img_size, 
               filters , 
               alpha = 0.3, 
               beta = 0.15 , 
               gamma = 1.0, 
               kappa = 0.5,
               data_format='channels_last', 
               name=None, **kwargs):
    super(ComplexSpectralMaxPool2D, self).__init__(name=name, **kwargs)
    if data_format is None:
      data_format = backend.image_data_format()
    self.data_format = conv_utils.normalize_data_format(data_format)
    self.input_spec = InputSpec(ndim=4)
    self.img_size = img_size
    self.filters = filters
    self.alpha = alpha
    self.beta = beta
    self.gamma = gamma
    self.kappa = kappa
 
  def call(self, inputs):
    im_fft = dtypes.complex(inputs[:,:,:,:self.filters], inputs[:,:,:,self.filters:])
    im_fft = transpose(im_fft, [0,3,1,2])
    
    filter_size = self._get_sp_dim(self.img_size)
    #filter_size = math.multiply(self.gamma, filter_size)
    im_transformed = self._spectral_pool(im_fft, filter_size)
    
    freq_dropout_lower_bound, freq_dropout_upper_bound = self._get_frq_dropout_bounds(filter_size, self.kappa)
    if (freq_dropout_lower_bound is not None and
        freq_dropout_upper_bound is not None):
        
        tf_random_cutoff = uniform([], freq_dropout_lower_bound, freq_dropout_upper_bound)
        dropout_mask = self._frequency_dropout_mask(
                            filter_size,
                            tf_random_cutoff)
        
        im_downsampled = im_transformed * dropout_mask
    else:
        im_downsampled = im_transformed
    
    outputs_downsampled = transpose(im_downsampled, [0,2,3,1])
    outputs = concat([math.real(outputs_downsampled), math.imag(outputs_downsampled)], axis = -1)
    
    return outputs
    
    
  def _spectral_pool(self, images, filter_size):
    assert len(images.get_shape().as_list()) == 4
    assert filter_size >= 3

    if filter_size % 2 == 1:
        n = int((filter_size-1)/2)
        top_left = images[:, :, :n+1, :n+1]
        top_right = images[:, :, :n+1, -n:]
        bottom_left = images[:, :, -n:, :n+1]
        bottom_right = images[:, :, -n:, -n:]
        top_combined = concat([top_left, top_right], axis=-1)
        bottom_combined = concat([bottom_left, bottom_right], axis=-1)
        all_together = concat([top_combined, bottom_combined], axis=-2)
    else:
        n = filter_size // 2
        top_left = images[:, :, :n, :n]
        top_middle = expand_dims(
            cast(0.5 ** 0.5, complex64) *
            (images[:, :, :n, n] + images[:, :, :n, -n]),
            -1)
        top_right = images[:, :, :n, -(n-1):]
        middle_left = expand_dims(
            cast(0.5 ** 0.5, complex64) *
            (images[:, :, n, :n] + images[:, :, -n, :n]), 
            -2)
        middle_middle = expand_dims(
            expand_dims(
                cast(0.5, complex64) *
                (images[:, :, n, n] + images[:, :, n, -n] +
                images[:, :, -n, n] + images[:, :, -n, -n]),
                -1), -1)
        middle_right = expand_dims(
            cast(0.5 ** 0.5, complex64) *
            (images[:, :, n, -(n-1):] + images[:, :, -n, -(n-1):]),
            -2)
        bottom_left = images[:, :, -(n-1):, :n]
        bottom_middle = expand_dims(
            cast(0.5 ** 0.5, complex64) *
            (images[:, :, -(n-1):, n] + images[:, :, -(n-1):, -n]),
            -1)
        bottom_right = images[:, :, -(n-1):, -(n-1):]
        top_combined = concat(
            [top_left, top_middle, top_right],
            axis=-1)
        middle_combined = concat(
            [middle_left, middle_middle, middle_right],
            axis=-1)
        bottom_combined = concat(
            [bottom_left, bottom_middle, bottom_right],
            axis=-1)
        all_together = concat(
            [top_combined, middle_combined, bottom_combined],
            axis=-2)
    return all_together
    
  def _get_frq_dropout_bounds(self, n, m):
    c = self.alpha + m* (self.beta - self.alpha)
    freq_dropout_lower_bound = c * (1. + n // 2)
    freq_dropout_upper_bound = (1. + n // 2)

    return freq_dropout_lower_bound, freq_dropout_upper_bound

  def _get_sp_dim(self, n):
        fsize = int(self.gamma * n)
        # minimum size is 3:
        return max(3, fsize)

  def _frequency_dropout_mask(self, height, frequency_to_truncate_above):
    cutoff_shape = frequency_to_truncate_above.get_shape().as_list()
    assert len(cutoff_shape) == 0

    mid = height // 2
    if height % 2 == 1:
        go_to = mid + 1
    else:
        go_to = mid
    indexes = np.concatenate((
        np.arange(go_to),
        np.arange(mid, 0, -1)
    )).astype(np.float32)

    xs = np.broadcast_to(indexes, (height, height))
    ys = np.broadcast_to(np.expand_dims(indexes, -1), (height, height))
    highest_frequency = np.maximum(xs, ys)

    comparison_mask = constant(highest_frequency)
    dropout_mask = cast(less_equal(
        comparison_mask,
        frequency_to_truncate_above
    ), complex64)
    
    return dropout_mask

  def get_config(self):
    config = {
        'img_size': self.img_size,
        'filters':self.filters,
        'alpha': self.alpha,
        'beta': self.beta,
        'gamma': self.gamma,
        'kappa': self.kappa,
        'data_format': self.data_format,
        'name': self.name
    }
    base_config = super(ComplexSpectralMaxPool2D, self).get_config()
    return dict(list(base_config.items()) + list(config.items()))



import os
import sys
import numpy as np
import tensorflow as tf

#Utility helper functions: custom learning rate scheduler, custom optimizers and other custom 
#functions can be defined here

def load_trained_model(model_type = 'fcunet'):
    model_path = '../model_weights/fcunet_strain_mapping_weights_latest/'
    print('loading latest trained disk detection model {} \n'.format(model_path))
    model = tf.keras.models.load_model(model_path,
                                       custom_objects={'lrScheduler': lrScheduler(128)})
    return model

class lrScheduler(tf.keras.optimizers.schedules.LearningRateSchedule):
    def __init__(self, d_model, alpha =10, warmup_steps=50, name=None):
        super(lrScheduler, self).__init__()

        self.dim_model = d_model
        self.d_model = tf.cast(self.dim_model, tf.float32)
        self.alpha = alpha
        self.warmup_steps = warmup_steps
        if name is None:
            self.name = 'lr_scheduler'
        else:
            self.name = name

    def __call__(self, step):
        arg1 = tf.math.rsqrt(step)
        arg2 = step * (self.warmup_steps ** -1.5)

        return tf.math.rsqrt(self.alpha*self.d_model) * tf.math.minimum(arg1, arg2)
            
    def get_config(self):
        return {"d_model": self.dim_model,
                "alpha": self.alpha,
                "warmup_steps": self.warmup_steps,
                "name": self.name
                }


import os
import sys
import h5py
import numpy as np
from tqdm import tqdm
import tensorflow as tf

#Custom libraries - User defined
from ..layers import convSpectral
from ..layers import poolSpectral

'''
These functions are to create the custom layers
'''

# %%
#Adapted from get_cross_correlation algorithm in py4DStem (corrPower = 0)

def cross_correlate(x):
    cb = x[0]
    pr = x[1]
    cb = tf.keras.backend.permute_dimensions(cb, (3,0,1,2))
    pr = tf.keras.backend.permute_dimensions(pr, (3,0,1,2))
    
    #shift probe to the origin
    pr = tf.signal.ifftshift(pr, axes = (2,3))
    cb = tf.signal.ifftshift(cb, axes = (2,3))
    cbed = tf.signal.fft2d(tf.cast(cb,tf.complex64))
    probe = tf.signal.fft2d(tf.cast(pr,tf.complex64))
    
    ccff = tf.multiply(cbed, tf.math.conj(probe))
    cc = tf.math.real(tf.signal.fftshift(tf.signal.ifft2d(ccff)))
    cc = tf.keras.backend.permute_dimensions(cc, (1,2,3,0))
    
    #normalize each cross-corr
    cc, _ = tf.linalg.normalize(cc, axis = (1,2))                 
        
    return cc


def cross_correlate_ff(x):
    cb = x[0]
    pr = x[1]
    cb = tf.keras.backend.permute_dimensions(cb, (3,0,1,2))
    pr = tf.keras.backend.permute_dimensions(pr, (3,0,1,2))
    
    #shift probe to the origin
    pr = tf.signal.ifftshift(pr, axes = (2,3))
    cb = tf.signal.ifftshift(cb, axes = (2,3))
    cbed = tf.signal.fft2d(tf.cast(cb,tf.complex64))
    probe = tf.signal.fft2d(tf.cast(pr,tf.complex64))
    
    ccff = tf.multiply(cbed, tf.math.conj(probe))
    #cc = tf.math.real(tf.signal.ifft2d(ccff))
    ccff = tf.keras.backend.permute_dimensions(ccff, (1,2,3,0))
    
    #normalize each cross-corr
    ccff, _ = tf.linalg.normalize(ccff, axis = (1,2)) 
    ccff_real = tf.math.real(ccff)
    ccff_im = tf.math.imag(ccff)
    ccff = tf.concat([ccff_real,ccff_im], axis = -1)
    
    return ccff


def cross_correlate_iff(x):
    input_channel = x.shape[-1] // 2

    input_complex = tf.dtypes.complex(x[:,:,:,:input_channel], x[:,:,:,input_channel:])
    input_transposed = tf.transpose(input_complex, [0,3,1,2])
    #output_complex = tf.math.real(tf.signal.ifft2d(input_transposed))
    output_complex = tf.math.real(tf.signal.fftshift(tf.signal.ifft2d(input_transposed), axes=(2,3)))
    output = tf.transpose(output_complex, [0,2,3,1])
    
    return output

# %%
#MCDropout
class MonteCarloDropout(tf.keras.layers.Dropout):
    def call(self, inputs):
        return super().call(inputs, training=True)

def conv2d_block(input_tensor, n_filters, n_depth=2, input_channel=1, kernel_size = 3, activation = 'relu', dp_rate = 0.1, batchnorm = True):
    """Function to add n (depth) convolutional layers with the parameters passed to it"""
    
    kl = 'he_uniform'
    if batchnorm:
        x = tf.keras.layers.BatchNormalization()(input_tensor)
    else:
        x = input_tensor
    
    for _ in range(n_depth):
        x = tf.keras.layers.Conv2D(filters = n_filters, kernel_size = (kernel_size, kernel_size),             
                               padding = 'same', kernel_initializer = kl, bias_initializer = kl, kernel_regularizer= 'l2', bias_regularizer= 'l2')(x)
        if batchnorm:
            x = tf.keras.layers.BatchNormalization()(x)
        x = tf.keras.layers.Activation(activation)(x)
        x = MonteCarloDropout(dp_rate)(x)
    
    return x

def convSpec2d_block(input_tensor, n_filters, n_depth=2, input_channel=1, kernel_size = 3, activation = 'relu', dp_rate = 0.1, batchnorm = True):
    """Function to add n (depth) complex convolutional layers with the parameters passed to it"""
    # Adapt Oren Rippel's paper (https://github.com/oracleofnj/spectral-repr-cnns/tree/master/src/modules)
    
    kl = 'he_uniform'
    if batchnorm:
        x = tf.keras.layers.BatchNormalization()(input_tensor)
    else:
        x = input_tensor
            
    for _ in range(n_depth):
        x = convSpectral.ConvComplex2D(rank=2, filters = n_filters, kernel_size = (kernel_size, kernel_size),
                                padding = 'same', kernel_initializer = kl, bias_initializer = kl, kernel_regularizer= 'l2', bias_regularizer= 'l2')(x)
        if batchnorm:
            x = tf.keras.layers.BatchNormalization()(x)
        x = tf.keras.layers.Activation(activation)(x)
        x = MonteCarloDropout(dp_rate)(x)
    
    return x


def convSpectral2d_block(input_tensor, n_filters, n_depth=2, input_channel=1, kernel_size = 3, activation = 'relu', dp_rate = 0.1, batchnorm = True):
    """Function to add n (depth) spectral convolutional layers with the parameters passed to it"""
    # Adapt Oren Rippel's paper (https://github.com/oracleofnj/spectral-repr-cnns/tree/master/src/modules)
    
    kl = 'he_uniform'
    if batchnorm:
        x = tf.keras.layers.BatchNormalization()(input_tensor)
    else:
        x = input_tensor
            
    for _ in range(n_depth):
        x = convSpectral.ConvSpectral2D(rank=2, filters = n_filters, kernel_size = (kernel_size, kernel_size),
                                padding = 'same', kernel_initializer = kl, bias_initializer = kl, kernel_regularizer= 'l2', bias_regularizer= 'l2')(x)
        if batchnorm:
            x = tf.keras.layers.BatchNormalization()(x)
        x = tf.keras.layers.Activation(activation)(x)
        x = MonteCarloDropout(dp_rate)(x)
    
    return x



"""
This part is taken from tensorflow addons utils
"""

"""Types for typing functions signatures."""

from typing import Union, Callable, List

import numpy as np
import tensorflow as tf

Number = Union[
    float,
    int,
    np.float16,
    np.float32,
    np.float64,
    np.int8,
    np.int16,
    np.int32,
    np.int64,
    np.uint8,
    np.uint16,
    np.uint32,
    np.uint64,
]

Initializer = Union[None, dict, str, Callable]
Regularizer = Union[None, dict, str, Callable]
Constraint = Union[None, dict, str, Callable]
Activation = Union[None, str, Callable]
Optimizer = Union[tf.keras.optimizers.Optimizer, str]

TensorLike = Union[
    List[Union[Number, list]],
    tuple,
    Number,
    np.ndarray,
    tf.Tensor,
    tf.SparseTensor,
    tf.Variable,
]
FloatTensorLike = Union[tf.Tensor, float, np.float16, np.float32, np.float64]
AcceptableDTypes = Union[tf.DType, np.dtype, type, int, str, None]


#!/usr/bin/env python
# coding: utf-8
# %%
#Copyright: Joydeep Munshi
#Date: 07/01/2021

import os
import sys
import time
import py4DSTEM
import numpy as np
from tqdm import tqdm

from itertools import product
from functools import reduce
from operator import mul
from numpy import integer

from collections.abc import Iterator

class nditer(Iterator):
    def __init__(self,*args):
        if len(args) > 1:
            self._it = product(*args)
        else:
            self._it = args[0]
        self._l = reduce(mul,[a.__len__() for a in args])
    def __iter__(self):
        return self._it.__iter__()
    def __next__(self):
        return self._it.__next__()
    def __len__(self):
        return self._l

def tqdmnd(*args,**kwargs):
    r = [range(i) if isinstance(i,(int,integer)) else i for i in args]
    return tqdm(nditer(*r),**kwargs)

def fourier_resample(
    array, 
    scale=None, 
    output_size=None,
    force_nonnegative=False,
    bandlimit_nyquist=None,
    bandlimit_power=2, 
    dtype=np.float32):
    """
    copied Colin Ophus's code from py4dstem
    
    Resize a 2D array along any dimension, using Fourier interpolation / extrapolation.
    For 4D input arrays, only the final two axes can be resized.
    The scaling of the array can be specified by passing either `scale`, which sets
    the scaling factor along both axes to be scaled; or by passing `output_size`, 
    which specifies the final dimensions of the scaled axes (and allows for different
    scaling along the x,y or kx,ky axes.)
    Args:
        array (2D/4D numpy array): Input array, or 4D stack of arrays, to be resized. 
        scale (float): scalar value giving the scaling factor for all dimensions
        output_size (2-tuple of ints): two values giving either the (x,y) output size for 2D, or (kx,ky) for 4D
        force_nonnegative (bool): Force all outputs to be nonnegative, after filtering
        bandlimit_nyquist (float): Gaussian filter information limit in Nyquist units (0.5 max in both directions)
        bandlimit_power (float): Gaussian filter power law scaling (higher is sharper)
        dtype (numpy dtype): datatype for binned array. default is single precision float
    Returns:
        the resized array (2D/4D numpy array)
    """

    # Verify input is 2D or 4D
    if np.size(array.shape) != 2 and np.size(array.shape) != 4:
        raise Exception('Function does not support arrays with ' \
         + str(np.size(array.shape)) + ' dimensions')

    # Get input size from last 2 dimensions
    input__size = array.shape[-2:]


    if scale is not None:
        assert output_size is None, 'Cannot specify both a scaling factor and output size'
        assert np.size(scale) == 1, 'scale should be a single value'
        scale = np.asarray(scale)
        output_size = (input__size * scale).astype('intp')
    else:
        assert scale is None, 'Cannot specify both a scaling factor and output size'
        assert np.size(output_size) == 2, 'output_size must contain two values'
        output_size = np.asarray(output_size)
    
    scale_output = np.prod(output_size) / np.prod(input__size)


    if bandlimit_nyquist is not None:
        kx = np.fft.fftfreq(output_size[0])
        ky = np.fft.fftfreq(output_size[1])
        k2 = kx[:,None]**2 + ky[None,:]**2
        # Gaussian filter 
        k_filt = np.exp((k2**(bandlimit_power/2))/(-2*bandlimit_nyquist**bandlimit_power))


    # generate slices
    # named as {dimension}_{corner}_{in_/out},
    # where corner is ul, ur, ll, lr for {upper/lower}{left/right}

    # x slices
    if output_size[0] > input__size[0]:
        # x dimension increases
        x0 = int((input__size[0]+1)//2)
        x1 = int( input__size[0]   //2)

        x_ul_out = slice(0, x0)
        x_ul_in_ = slice(0, x0)

        x_ll_out = slice(0-x1+output_size[0], output_size[0])
        x_ll_in_ = slice(0-x1+input__size[0], input__size[0])

        x_ur_out = slice(0, x0)
        x_ur_in_ = slice(0, x0)

        x_lr_out = slice(0-x1+output_size[0], output_size[0])
        x_lr_in_ = slice(0-x1+input__size[0], input__size[0])

    elif output_size[0] < input__size[0]:
        # x dimension decreases
        x0 = int((output_size[0]+1)//2)
        x1 = int( output_size[0]   //2)

        x_ul_out = slice(0, x0)
        x_ul_in_ = slice(0, x0)

        x_ll_out = slice(0-x1+output_size[0], output_size[0])
        x_ll_in_ = slice(0-x1+input__size[0], input__size[0])

        x_ur_out = slice(0, x0)
        x_ur_in_ = slice(0, x0)

        x_lr_out = slice(0-x1+output_size[0], output_size[0])
        x_lr_in_ = slice(0-x1+input__size[0], input__size[0])

    else:
        # x dimension does not change
        x_ul_out = slice(None)
        x_ul_in_ = slice(None)

        x_ll_out = slice(None)
        x_ll_in_ = slice(None)

        x_ur_out = slice(None)
        x_ur_in_ = slice(None)

        x_lr_out = slice(None)
        x_lr_in_ = slice(None)

    #y slices
    if output_size[1] > input__size[1]:
        # y increases
        y0 = int((input__size[1]+1)//2)
        y1 = int( input__size[1]   //2)

        y_ul_out = slice(0, y0)
        y_ul_in_ = slice(0, y0)

        y_ll_out = slice(0, y0)
        y_ll_in_ = slice(0, y0)

        y_ur_out = slice(0-y1+output_size[1], output_size[1])
        y_ur_in_ = slice(0-y1+input__size[1], input__size[1])

        y_lr_out = slice(0-y1+output_size[1], output_size[1])
        y_lr_in_ = slice(0-y1+input__size[1], input__size[1])

    elif output_size[1] < input__size[1]:
        # y decreases
        y0 = int((output_size[1]+1)//2)
        y1 = int( output_size[1]   //2)

        y_ul_out = slice(0, y0)
        y_ul_in_ = slice(0, y0)

        y_ll_out = slice(0, y0)
        y_ll_in_ = slice(0, y0)

        y_ur_out = slice(0-y1+output_size[1], output_size[1])
        y_ur_in_ = slice(0-y1+input__size[1], input__size[1])

        y_lr_out = slice(0-y1+output_size[1], output_size[1])
        y_lr_in_ = slice(0-y1+input__size[1], input__size[1])

    else:
        # y dimension does not change
        y_ul_out = slice(None)
        y_ul_in_ = slice(None)

        y_ll_out = slice(None)
        y_ll_in_ = slice(None)

        y_ur_out = slice(None)
        y_ur_in_ = slice(None)

        y_lr_out = slice(None)
        y_lr_in_ = slice(None)

    if len(array.shape) == 2:
        # image array        
        array_resize = np.zeros(output_size, dtype=np.complex64)
        array_fft = np.fft.fft2(array)

        # copy each quadrant into the resize array
        array_resize[x_ul_out, y_ul_out] = array_fft[x_ul_in_, y_ul_in_]
        array_resize[x_ll_out, y_ll_out] = array_fft[x_ll_in_, y_ll_in_]
        array_resize[x_ur_out, y_ur_out] = array_fft[x_ur_in_, y_ur_in_]
        array_resize[x_lr_out, y_lr_out] = array_fft[x_lr_in_, y_lr_in_]

        # Band limit if needed
        if bandlimit_nyquist is not None:
            array_resize *= k_filt

        # Back to real space
        array_resize = np.real(np.fft.ifft2(array_resize)).astype(dtype)


    elif len(array.shape) == 4:
        # This case is the same as the 2D case, but loops over the probe index arrays

        # init arrays
        array_resize = np.zeros((*array.shape[:2], *output_size), dtype)
        array_fft = np.zeros(input__size, dtype=np.complex64)
        array_output = np.zeros(output_size, dtype=np.complex64)

        for (Rx,Ry) in tqdmnd(array.shape[0],array.shape[1],desc='Resampling 4D datacube',unit='DP',unit_scale=True):
            array_fft[:,:] = np.fft.fft2(array[Rx,Ry,:,:])
            array_output[:,:] = 0

            # copy each quadrant into the resize array
            array_output[x_ul_out,y_ul_out] = array_fft[x_ul_in_,y_ul_in_]
            array_output[x_ll_out,y_ll_out] = array_fft[x_ll_in_,y_ll_in_]
            array_output[x_ur_out,y_ur_out] = array_fft[x_ur_in_,y_ur_in_]
            array_output[x_lr_out,y_lr_out] = array_fft[x_lr_in_,y_lr_in_]

            # Band limit if needed
            if bandlimit_nyquist is not None:
                array_output *= k_filt

            # Back to real space
            array_resize[Rx,Ry,:,:] = np.real(np.fft.ifft2(array_output)).astype(dtype)

    # Enforce positivity if needed, after filtering
    if force_nonnegative:
        array_resize = np.maximum(array_resize,0)
        
    # Normalization
    array_resize = array_resize * scale_output

    return array_resize

class py4dstemModel(object):
    def __init__(self, maxNumPeaks=100, relativeToPeak=0, probe_kernel_type='gaussian'):
        self.maxNumPeaks=maxNumPeaks
        self.relativeToPeak = relativeToPeak
        self.probe_kernel_type = probe_kernel_type
        
    def integrate_disks(self, image, maxima_x,maxima_y,maxima_int,thresold=1):
        disks = []
        for x,y,i in zip(maxima_x,maxima_y,maxima_int):
            disk = image[int(x)-thresold:int(x)+thresold, int(y)-thresold:int(y)+thresold]
            disks.append(np.average(disk))
        disks = disks/max(disks)
        return (maxima_x,maxima_y,disks)

    def py4dstemDiskDet(self, probe, cbed, 
                        p4_param_dict):
        if p4_param_dict['probe_kernel_type']:
            self.probe_kernel_type = p4_param_dict['probe_kernel_type']
        if self.probe_kernel_type == 'gaussian':
            probe_kernel = py4DSTEM.process.diskdetection.probe.get_probe_kernel_edge_gaussian(probe, sigma_probe_scale=1)
        elif self.probe_kernel_type == 'sigmoid_sine':
            probe_kernel = py4DSTEM.process.diskdetection.get_probe_kernel_edge_sigmoid(probe,
                                                                            p4_param_dict['sigmoid_min'],
                                                                            p4_param_dict['sigmoid_max'],
                                                                            type='sine_squared')
        elif self.probe_kernel_type == 'sigmoid_logistic':
            probe_kernel = py4DSTEM.process.diskdetection.get_probe_kernel_edge_sigmoid(probe,
                                                                            p4_param_dict['sigmoid_min'],
                                                                            p4_param_dict['sigmoid_max'],
                                                                            type='logistic')
        else:
            print('Kernel type is not yet implemented \n')
        
        peaks = py4DSTEM.process.diskdetection.find_Bragg_disks_single_DP(cbed, probe_kernel,
                                                                          corrPower = p4_param_dict['corrPower'],
                                                                          sigma = p4_param_dict['sigma'],
                                                                          edgeBoundary = p4_param_dict['edgeBoundary'],
                                                                          minRelativeIntensity = p4_param_dict['minRelativeIntensity'],
                                                                          relativeToPeak = self.relativeToPeak,
                                                                          minPeakSpacing = p4_param_dict['minPeakSpacing'],
                                                                          maxNumPeaks = self.maxNumPeaks,
                                                                          subpixel = p4_param_dict['subpixel'],
                                                                          upsample_factor = p4_param_dict['upsample_factor'])
        return peaks

    def aimlDiskDet(self, image, 
                    aiml_param_dict,
                    integrate=True, 
                    thresold=1):
        maxima_x,maxima_y,maxima_int = py4DSTEM.process.utils.get_maxima_2D(image, 
                                                                            minRelativeIntensity=aiml_param_dict['minRelativeIntensity'], 
                                                                            edgeBoundary=aiml_param_dict['edgeBoundary'],
                                                                            maxNumPeaks=self.maxNumPeaks,
                                                                            minSpacing = aiml_param_dict['minPeakSpacing'],
                                                                            subpixel=aiml_param_dict['subpixel'],
                                                                            upsample_factor=aiml_param_dict['upsample_factor'])
            
        if integrate:
            maxima_x, maxima_y, maxima_int = self.integrate_disks(image, maxima_x,maxima_y,maxima_int,thresold=thresold)
            
        coords = [('qx',float),('qy',float),('intensity',float)]
        peaks = py4DSTEM.io.PointList(coordinates=coords)
        peaks.add_tuple_of_nparrays((maxima_x, maxima_y, maxima_int))
        
        return peaks

    def score(self, ground_truth, 
              prediction, 
              tr_image=None, 
              pred_image=None, 
              cutoff=0.05, 
              pixel_size = 0.1,
              integrate_disk=False):
        if integrate_disk:
            true_x, true_y, true_int = self.integrate_disks(tr_image, ground_truth[0], ground_truth[1], ground_truth[2])
            pred_x, pred_y, pred_int = self.integrate_disks(pred_image, prediction[0], prediction[1], prediction[2])
        else:
            true_x, true_y, true_int = ground_truth[0], ground_truth[1], ground_truth[2]/max(ground_truth[2])
            pred_x, pred_y, pred_int = prediction[0], prediction[1], prediction[2]/max(prediction[2])
    
        true_coord = np.asarray((true_x,true_y)).T
        pred_coord = np.asarray((pred_x,pred_y)).T
        true_coord = np.delete(true_coord, np.argmax(true_int), axis=0)
        pred_coord = np.delete(pred_coord, np.argmax(pred_int), axis=0)
        true_int_ = np.delete(true_int, np.argmax(true_int), axis=0)
        pred_int_ = np.delete(pred_int, np.argmax(pred_int), axis=0)
        closest_true = self.find_closest_disks(true_coord, pred_coord)
        closest_pred = self.find_closest_disks(pred_coord, true_coord)
        dist_true = np.sum((true_coord - closest_true)**2, axis=1)
        dist_pred = np.sum((pred_coord - closest_pred)**2, axis=1)
        sub_true = dist_true <= (cutoff/pixel_size)**2
        sub_pred = dist_pred <= (cutoff/pixel_size)**2
        #print(np.sum(sub_true),np.sum(sub_pred))
        #assert(np.sum(sub_true) == np.sum(sub_pred)), "True positive counts from measured (predicted) and ground truth are different; something is wrong!!"
    
        #Intensity weighted
        TP = np.sum(sub_pred)
        FN = np.sum(np.logical_not(sub_true))
        FP = np.sum(np.logical_not(sub_pred))
        accuracy = TP/(TP+FP+FN)
        
        TP_int = np.sum(pred_int_[sub_pred])/ np.sum(pred_int_)
        FN_int = np.sum(true_int_[np.logical_not(sub_true)])/ np.sum(true_int_)
        FP_int = np.sum(pred_int_[np.logical_not(sub_pred)])/ np.sum(pred_int_)
        accuracy_int = TP_int/(TP_int+FP_int+FN_int)
    
        return accuracy, accuracy_int, (true_x, true_y, true_int), (pred_x, pred_y, pred_int)
    
    def find_closest_disks(self, coord1, coord2):
        closest = [] 
        for coord in range(coord1.shape[0]):
            dist_sum = np.sum((coord2 - coord1[coord])**2, axis=1)
            closest.append(coord2[np.argmin(dist_sum),:])
        return np.asarray(closest)
    
    def calculate_fom(self, data, params):
        assert(data['probe'].shape == data['cbed'].shape)
        assert(data['probe'].shape == data['vg'].shape)
        image_num = data['probe'].shape[0]
        height = data['probe'].shape[1]
        width = data['probe'].shape[2]
        channel = data['probe'].shape[3]
    
        accuracy_int = 0.0
        for i in range(image_num):
            peaks = self.py4dstemDiskDet(data['probe'][i,:,:,0], data['cbed'][i,:,:,0], 
                                    params['corrPower'], 
                                    params['sigma'], 
                                    params['edgeBoundary'], 
                                    params['minRelativeIntensity'],  
                                    params['minPeakSpacing'], 
                                    'multicorr', 
                                    params['upsample_factor'],
                                    params['sigmoid_min'],
                                    params['sigmoid_max'])
            maxima_x, maxima_y, maxima_int =  self.aimlDiskDet(data['vg'][i,:,:,0], 
                                                          minRelativeIntensity = 0.05, 
                                                            subpixel='multicorr',
                                                            integrate = True)
        
            ground_truth_disk = [maxima_x, maxima_y, maxima_int]
            py4dstem_disk = [peaks.data['qx'],peaks.data['qy'],peaks.data['intensity']]
            accuracy_int += self.score(ground_truth_disk, py4dstem_disk)
        
        return accuracy_int/float(image_num)  


from .complexUnet import *
from .complexSpecPoolUnet import *
from .complexUnetMultiHead import *
from .spectralPoolUnet import *
from .spectralUnet import *
from .unet import *
from .spectralPoolDenseNet import *


"""
This part is taken from tensorflow addons image module. Available from tensorflow 2.3.0 and beyond.
"""

"""Image warping using per-pixel flow vectors."""

import numpy as np
import tensorflow as tf

from ..utils import aug_utils
from typing import Optional


@tf.function
def interpolate_bilinear(
    grid: aug_utils.TensorLike,
    query_points: aug_utils.TensorLike,
    indexing: str = "ij",
    name: Optional[str] = None,
):
    """Similar to Matlab's interp2 function.
    Finds values for query points on a grid using bilinear interpolation.
    Args:
      grid: a 4-D float `Tensor` of shape `[batch, height, width, channels]`.
      query_points: a 3-D float `Tensor` of N points with shape
        `[batch, N, 2]`.
      indexing: whether the query points are specified as row and column (ij),
        or Cartesian coordinates (xy).
      name: a name for the operation (optional).
    Returns:
      values: a 3-D `Tensor` with shape `[batch, N, channels]`
    Raises:
      ValueError: if the indexing mode is invalid, or if the shape of the
        inputs invalid.
    """
    if indexing != "ij" and indexing != "xy":
        raise ValueError("Indexing mode must be 'ij' or 'xy'")

    with tf.name_scope(name or "interpolate_bilinear"):
        grid = tf.convert_to_tensor(grid)
        query_points = tf.convert_to_tensor(query_points)

        # grid shape checks
        grid_static_shape = grid.shape
        grid_shape = tf.shape(grid)
        if grid_static_shape.dims is not None:
            if len(grid_static_shape) != 4:
                raise ValueError("Grid must be 4D Tensor")
            if grid_static_shape[1] is not None and grid_static_shape[1] < 2:
                raise ValueError("Grid height must be at least 2.")
            if grid_static_shape[2] is not None and grid_static_shape[2] < 2:
                raise ValueError("Grid width must be at least 2.")
        else:
            with tf.control_dependencies(
                [
                    tf.debugging.assert_greater_equal(
                        grid_shape[1], 2, message="Grid height must be at least 2."
                    ),
                    tf.debugging.assert_greater_equal(
                        grid_shape[2], 2, message="Grid width must be at least 2."
                    ),
                    tf.debugging.assert_less_equal(
                        tf.cast(
                            grid_shape[0] * grid_shape[1] * grid_shape[2],
                            dtype=tf.dtypes.float32,
                        ),
                        np.iinfo(np.int32).max / 8.0,
                        message="The image size or batch size is sufficiently "
                        "large that the linearized addresses used by "
                        "tf.gather may exceed the int32 limit.",
                    ),
                ]
            ):
                pass

        # query_points shape checks
        query_static_shape = query_points.shape
        query_shape = tf.shape(query_points)
        if query_static_shape.dims is not None:
            if len(query_static_shape) != 3:
                raise ValueError("Query points must be 3 dimensional.")
            query_hw = query_static_shape[2]
            if query_hw is not None and query_hw != 2:
                raise ValueError("Query points last dimension must be 2.")
        else:
            with tf.control_dependencies(
                [
                    tf.debugging.assert_equal(
                        query_shape[2],
                        2,
                        message="Query points last dimension must be 2.",
                    )
                ]
            ):
                pass

        batch_size, height, width, channels = (
            grid_shape[0],
            grid_shape[1],
            grid_shape[2],
            grid_shape[3],
        )

        num_queries = query_shape[1]

        query_type = query_points.dtype
        grid_type = grid.dtype

        alphas = []
        floors = []
        ceils = []
        index_order = [0, 1] if indexing == "ij" else [1, 0]
        unstacked_query_points = tf.unstack(query_points, axis=2, num=2)

        for i, dim in enumerate(index_order):
            with tf.name_scope("dim-" + str(dim)):
                queries = unstacked_query_points[dim]

                size_in_indexing_dimension = grid_shape[i + 1]

                # max_floor is size_in_indexing_dimension - 2 so that max_floor + 1
                # is still a valid index into the grid.
                max_floor = tf.cast(size_in_indexing_dimension - 2, query_type)
                min_floor = tf.constant(0.0, dtype=query_type)
                floor = tf.math.minimum(
                    tf.math.maximum(min_floor, tf.math.floor(queries)), max_floor
                )
                int_floor = tf.cast(floor, tf.dtypes.int32)
                floors.append(int_floor)
                ceil = int_floor + 1
                ceils.append(ceil)

                # alpha has the same type as the grid, as we will directly use alpha
                # when taking linear combinations of pixel values from the image.
                alpha = tf.cast(queries - floor, grid_type)
                min_alpha = tf.constant(0.0, dtype=grid_type)
                max_alpha = tf.constant(1.0, dtype=grid_type)
                alpha = tf.math.minimum(tf.math.maximum(min_alpha, alpha), max_alpha)

                # Expand alpha to [b, n, 1] so we can use broadcasting
                # (since the alpha values don't depend on the channel).
                alpha = tf.expand_dims(alpha, 2)
                alphas.append(alpha)

            flattened_grid = tf.reshape(grid, [batch_size * height * width, channels])
            batch_offsets = tf.reshape(
                tf.range(batch_size) * height * width, [batch_size, 1]
            )

        # This wraps tf.gather. We reshape the image data such that the
        # batch, y, and x coordinates are pulled into the first dimension.
        # Then we gather. Finally, we reshape the output back. It's possible this
        # code would be made simpler by using tf.gather_nd.
        def gather(y_coords, x_coords, name):
            with tf.name_scope("gather-" + name):
                linear_coordinates = batch_offsets + y_coords * width + x_coords
                gathered_values = tf.gather(flattened_grid, linear_coordinates)
                return tf.reshape(gathered_values, [batch_size, num_queries, channels])

        # grab the pixel values in the 4 corners around each query point
        top_left = gather(floors[0], floors[1], "top_left")
        top_right = gather(floors[0], ceils[1], "top_right")
        bottom_left = gather(ceils[0], floors[1], "bottom_left")
        bottom_right = gather(ceils[0], ceils[1], "bottom_right")

        # now, do the actual interpolation
        with tf.name_scope("interpolate"):
            interp_top = alphas[1] * (top_right - top_left) + top_left
            interp_bottom = alphas[1] * (bottom_right - bottom_left) + bottom_left
            interp = alphas[0] * (interp_bottom - interp_top) + interp_top

        return interp


def _get_dim(x, idx):
    if x.shape.ndims is None:
        return tf.shape(x)[idx]
    return x.shape[idx] or tf.shape(x)[idx]


@tf.function
def dense_image_warp(
    image: aug_utils.TensorLike, flow: aug_utils.TensorLike, name: Optional[str] = None
):
    """Image warping using per-pixel flow vectors.
    Apply a non-linear warp to the image, where the warp is specified by a
    dense flow field of offset vectors that define the correspondences of
    pixel values in the output image back to locations in the source image.
    Specifically, the pixel value at `output[b, j, i, c]` is
    `images[b, j - flow[b, j, i, 0], i - flow[b, j, i, 1], c]`.
    The locations specified by this formula do not necessarily map to an int
    index. Therefore, the pixel value is obtained by bilinear
    interpolation of the 4 nearest pixels around
    `(b, j - flow[b, j, i, 0], i - flow[b, j, i, 1])`. For locations outside
    of the image, we use the nearest pixel values at the image boundary.
    NOTE: The definition of the flow field above is different from that
    of optical flow. This function expects the negative forward flow from
    output image to source image. Given two images `I_1` and `I_2` and the
    optical flow `F_12` from `I_1` to `I_2`, the image `I_1` can be
    reconstructed by `I_1_rec = dense_image_warp(I_2, -F_12)`.
    Args:
      image: 4-D float `Tensor` with shape `[batch, height, width, channels]`.
      flow: A 4-D float `Tensor` with shape `[batch, height, width, 2]`.
      name: A name for the operation (optional).
      Note that image and flow can be of type `tf.half`, `tf.float32`, or
      `tf.float64`, and do not necessarily have to be the same type.
    Returns:
      A 4-D float `Tensor` with shape`[batch, height, width, channels]`
        and same type as input image.
    Raises:
      ValueError: if `height < 2` or `width < 2` or the inputs have the wrong
        number of dimensions.
    """
    with tf.name_scope(name or "dense_image_warp"):
        image = tf.convert_to_tensor(image)
        flow = tf.convert_to_tensor(flow)
        batch_size, height, width, channels = (
            _get_dim(image, 0),
            _get_dim(image, 1),
            _get_dim(image, 2),
            _get_dim(image, 3),
        )

        # The flow is defined on the image grid. Turn the flow into a list of query
        # points in the grid space.
        grid_x, grid_y = tf.meshgrid(tf.range(width), tf.range(height))
        #grid_x = grid_x - tf.math.reduce_mean(grid_x)
        #grid_y = grid_y - tf.math.reduce_mean(grid_y)
        stacked_grid = tf.cast(tf.stack([grid_y, grid_x], axis=2), flow.dtype)
        batched_grid = tf.expand_dims(stacked_grid, axis=0)
        query_points_on_grid = batched_grid - flow
        query_points_flattened = tf.reshape(
            query_points_on_grid, [batch_size, height * width, 2]
        )
        
        # Compute values at the query points, then reshape the result back to the
        # image grid.
        interpolated = interpolate_bilinear(image, query_points_flattened)
        interpolated = tf.reshape(interpolated, [batch_size, height, width, channels])
        return interpolated


#Added through version 0.0.6 release
import sys
import numpy as np

import tensorflow as tf
import tensorflow_addons as tfa

#Custom libraries - User defined
from ..utils import conv_utils
from ..layers import poolSpectral
from ..utils.conv_utils import convSpectral2d_block, conv2d_block

def spectralPoolDenseNet(input_channel, 
                         image_size, 
                         filter_size, 
                         dense_size, 
                         n_depth, 
                         n_conv,
                         n_dense, 
                         dense_output=1, 
                         dp_rate = 0.1, 
                         act = 'relu'):
    #Added through version 0.0.6 release
    input_size = (image_size, image_size, input_channel)
    inputs = tf.keras.Input(shape=input_size)

    pool = inputs

    print("\n")
    print("Building the spectral Dense-net blocks ... \n")
    for block in range(n_conv):
        if image_size//(2**block) > 8:
            conv = convSpectral2d_block(pool,
                                        filter_size * (2**block),
                                        n_depth= n_depth,
                                        dp_rate= dp_rate,
                                        activation= act)
            pool = poolSpectral.SpectralMaxPool2D(img_size = image_size//(2**block))(conv)

    print("\n")
    print("Building the some fully connected layers ... \n")
    dense = tf.keras.layers.Flatten()(pool)
    
    for dense_layer in reversed(range(n_dense)):
        dense = tf.keras.layers.Dense(dense_size * (2**dense_layer),
                                     activation= act,
                                     kernel_regularizer='l2')(dense)
        dense = tf.keras.layers.Dropout(dp_rate)(dense)

    print("\n")
    print("Building the final output dense layer ... \n")
    output = tf.keras.layers.Dense(dense_output,
                                   activation='softmax',
                                   kernel_regularizer='l2',
                                   dtype='float32')(dense)

    model = tf.keras.Model(inputs=inputs, outputs=output)

    return model


import functools
import six

from tensorflow import reshape
from tensorflow import einsum
from tensorflow import convert_to_tensor

from tensorflow.math import reduce_sum

from tensorflow.python.eager import context
from tensorflow.python.framework import tensor_shape
from tensorflow.python.keras import activations
from tensorflow.python.keras import backend
from tensorflow.python.keras import constraints
from tensorflow.python.keras import initializers
from tensorflow.python.keras import regularizers
from tensorflow.python.keras.engine.base_layer import Layer
from tensorflow.python.keras.engine.input_spec import InputSpec
# imports for backwards namespace compatibility
# pylint: disable=unused-import
# pylint: enable=unused-import
from tensorflow.python.keras.utils import conv_utils
from tensorflow.python.keras.utils import tf_utils
from tensorflow.python.ops import array_ops
from tensorflow.python.ops import nn
from tensorflow.python.ops import nn_ops
from tensorflow.python.util.tf_export import keras_export
# pylint: disable=g-classes-have-attributes
from tensorflow.image import extract_patches


class ConvQuad2D(Layer):
    def __init__(self,
                   rank,
                   filters,
                   kernel_size,
                   strides=(1,1),
                   padding='valid',
                   data_format=None,
                   dilation_rate=1,
                   groups=1,
                   activation=None,
                   use_bias=True,
                   kernel_initializer='glorot_uniform',
                   bias_initializer='zeros',
                   kernel_regularizer=None,
                   bias_regularizer=None,
                   activity_regularizer=None,
                   kernel_constraint=None,
                   bias_constraint=None,
                   trainable=True,
                   name=None,
                   conv_op=None,
                   **kwargs):
            super(ConvQuad2D, self).__init__(
                trainable=trainable,
                name=name,
                activity_regularizer=regularizers.get(activity_regularizer),
                **kwargs)
            self.rank = 2

            if isinstance(filters, float):
                filters = int(filters)
            self.filters = filters
            self.groups = groups or 1
            self.kernel_size = conv_utils.normalize_tuple(kernel_size, rank, 'kernel_size')
            self.strides = conv_utils.normalize_tuple(strides, rank, 'strides')
            self.padding = conv_utils.normalize_padding(padding)
            self.data_format = conv_utils.normalize_data_format(data_format)
            self.dilation_rate = conv_utils.normalize_tuple(
                dilation_rate, rank, 'dilation_rate')

            self.activation = activations.get(activation)
            self.use_bias = use_bias

            self.kernel_initializer = initializers.get(kernel_initializer)
            self.bias_initializer = initializers.get(bias_initializer)
            self.kernel_regularizer = regularizers.get(kernel_regularizer)
            self.bias_regularizer = regularizers.get(bias_regularizer)
            self.kernel_constraint = constraints.get(kernel_constraint)
            self.bias_constraint = constraints.get(bias_constraint)
            self.input_spec = InputSpec(min_ndim=self.rank + 2)

            self._validate_init()
            self._is_causal = self.padding == 'causal'
            self._channels_first = self.data_format == 'channels_first'
            self._tf_data_format = conv_utils.convert_data_format(
                self.data_format, self.rank + 2)
    
    
            #Quadtraic CNN initialization
    
            if isinstance(kernel_size, int):
                self.quad_kernel_size = conv_utils.normalize_tuple(
                    (kernel_size**2), rank, 'quad_kernel_size')
            else:
                self.quad_kernel_size = conv_utils.normalize_tuple(
                    (kernel_size[0]**2,kernel_size[1]**2), rank, 'quad_kernel_size')
    
    #Quadratic CNN using Volterra kernel theory
    def _volterra_conv(self, inputs, W, input_dim, quad_ksize, quad_strides, padding):
        input_patches = extract_patches(inputs,
                                        sizes= quad_ksize,
                                        strides=quad_strides,
                                        rates=quad_strides,
                                        padding=padding)
       
        input_patches_shape = (-1, 250, 250, self.kernel_size[0]*self.kernel_size[1], input_dim)
        #print(input_patches_shape)
        input_patches = array_ops.reshape(input_patches, input_patches_shape)
        V = einsum('abcid,abcjd,dijo->abcdo', input_patches, input_patches, W)
        return reduce_sum(V, 3)
    ##############################################

    def _validate_init(self):
            if self.filters is not None and self.filters % self.groups != 0:
                raise ValueError(
                      'The number of filters must be evenly divisible by the number of '
                      'groups. Received: groups={}, filters={}'.format(
                          self.groups, self.filters))

            if not all(self.kernel_size):
                raise ValueError('The argument `kernel_size` cannot contain 0(s). '
                                'Received: %s' % (self.kernel_size,))

            if (self.padding == 'causal' and not isinstance(self,(Conv1D, SeparableConv1D))):
                  raise ValueError('Causal padding is only supported for `Conv1D`'
                                   'and `SeparableConv1D`.')


    def build(self, input_shape):
        input_shape = tensor_shape.TensorShape(input_shape)
        input_channel = self._get_input_channel(input_shape)
        if input_channel % self.groups != 0:
            raise ValueError(
                  'The number of input channels must be evenly divisible by the number '
                  'of groups. Received groups={}, but the input has {} channels '
                  '(full input shape is {}).'.format(self.groups, input_channel,
                                                     input_shape))
        kernel_shape = self.kernel_size + (input_channel // self.groups, self.filters)

        self.kernel = self.add_weight(
            name='kernel',
            shape=kernel_shape,
            initializer=self.kernel_initializer,
            regularizer=self.kernel_regularizer,
            constraint=self.kernel_constraint,
            trainable=True,
            dtype=self.dtype)
    
        #Volterra kernel initialize
    
        self.quad_kernel_shape = (input_channel // self.groups,) + self.quad_kernel_size + (self.filters,)
    
        self.quad_kernel = self.add_weight(
            name='quad_kernel',
            shape=self.quad_kernel_shape,
            initializer=self.kernel_initializer,
            regularizer=self.kernel_regularizer,
            constraint=self.kernel_constraint,
            trainable=True,
            dtype=self.dtype)
    
        if self.use_bias:
            self.bias = self.add_weight(
                name='bias',
                shape=(self.filters,),
                initializer=self.bias_initializer,
                regularizer=self.bias_regularizer,
                constraint=self.bias_constraint,
                trainable=True,
                dtype=self.dtype)
        else:
            self.bias = None
        channel_axis = self._get_channel_axis()
        self.input_spec = InputSpec(min_ndim=self.rank + 2,
                                    axes={channel_axis: input_channel})

        # Convert Keras formats to TF native formats.
        if self.padding == 'causal':
            tf_padding = 'VALID'  # Causal padding handled in `call`.
        elif isinstance(self.padding, six.string_types):
            tf_padding = self.padding.upper()
        else:
            tf_padding = self.padding
        tf_dilations = list(self.dilation_rate)
        tf_strides = list(self.strides)

        tf_op_name = self.__class__.__name__
        if tf_op_name == 'Conv1D':
            tf_op_name = 'conv1d'  # Backwards compat.

        self._convolution_op = functools.partial(
            nn_ops.convolution_v2,
            strides=tf_strides,
            padding=tf_padding,
            dilations=tf_dilations,
            data_format=self._tf_data_format,
            name=tf_op_name)
    
        #Volterra quad CNN
        tf_quad_strides = list((1,) + self.strides + (1,))
        tf_quad_padding = tf_padding
        tf_quad_ksize = list((1,) +  self.kernel_size + (1,))
    
        self._quad_convolution_op = functools.partial(
            self._volterra_conv,
            input_dim = input_channel,
            quad_ksize = tf_quad_ksize,
            quad_strides=tf_quad_strides,
            padding=tf_quad_padding)
    
        self.built = True

    def call(self, inputs):
        if self._is_causal:  # Apply causal padding to inputs for Conv1D.
            inputs = array_ops.pad(inputs, self._compute_causal_padding(inputs))

        outputs = self._quad_convolution_op(inputs, self.quad_kernel) + self._convolution_op(inputs, self.kernel)

        if self.use_bias:
            output_rank = outputs.shape.rank
            if self.rank == 1 and self._channels_first:
                # nn.bias_add does not accept a 1D input tensor.
                bias = array_ops.reshape(self.bias, (1, self.filters, 1))
                outputs += bias
        else:
            # Handle multiple batch dimensions.
            if output_rank is not None and output_rank > 2 + self.rank:
                def _apply_fn(o):
                    return nn.bias_add(o, self.bias, data_format=self._tf_data_format)

                outputs = nn_ops.squeeze_batch_dims(outputs, _apply_fn, inner_rank=self.rank + 1)
            else:
                outputs = nn.bias_add(outputs, self.bias, data_format=self._tf_data_format)

        if self.activation is not None:
            return self.activation(outputs)
        return outputs


    def _spatial_output_shape(self, spatial_input_shape):
        return [
            conv_utils.conv_output_length(
                length,
                self.kernel_size[i],
                padding=self.padding,
                stride=self.strides[i],
                dilation=self.dilation_rate[i])
            for i, length in enumerate(spatial_input_shape)
            ]

    def compute_output_shape(self, input_shape):
        input_shape = tensor_shape.TensorShape(input_shape).as_list()
        batch_rank = len(input_shape) - self.rank - 1
        if self.data_format == 'channels_last':
            return tensor_shape.TensorShape(
              input_shape[:batch_rank]
              + self._spatial_output_shape(input_shape[batch_rank:-1])
              + [self.filters])
        else:
            return tensor_shape.TensorShape(
              input_shape[:batch_rank] + [self.filters] +
              self._spatial_output_shape(input_shape[batch_rank + 1:]))

    def _recreate_conv_op(self, inputs):  # pylint: disable=unused-argument
        return False

    def get_config(self):
        config = {
            'filters':
                self.filters,
            'kernel_size':
                self.kernel_size,
            'quad_kernel_size':
                self.quad_kernel_size,
            'strides':
                self.strides,
            'padding':
                self.padding,
            'data_format':
                self.data_format,
            'dilation_rate':
                self.dilation_rate,
            'groups':
                self.groups,
            'activation':
                activations.serialize(self.activation),
            'use_bias':
                self.use_bias,
            'kernel_initializer':
                initializers.serialize(self.kernel_initializer),
            'bias_initializer':
                initializers.serialize(self.bias_initializer),
            'kernel_regularizer':
                regularizers.serialize(self.kernel_regularizer),
            'bias_regularizer':
                regularizers.serialize(self.bias_regularizer),
            'activity_regularizer':
                regularizers.serialize(self.activity_regularizer),
            'kernel_constraint':
                constraints.serialize(self.kernel_constraint),
            'bias_constraint':
                constraints.serialize(self.bias_constraint)
        }
        base_config = super(ConvQuad2D, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

    def _compute_causal_padding(self, inputs):
        """Calculates padding for 'causal' option for 1-d conv layers."""
        left_pad = self.dilation_rate[0] * (self.kernel_size[0] - 1)
        if getattr(inputs.shape, 'ndims', None) is None:
            batch_rank = 1
        else:
            batch_rank = len(inputs.shape) - 2
        if self.data_format == 'channels_last':
            causal_padding = [[0, 0]] * batch_rank + [[left_pad, 0], [0, 0]]
        else:
            causal_padding = [[0, 0]] * batch_rank + [[0, 0], [left_pad, 0]]
        return causal_padding

    def _get_channel_axis(self):
        if self.data_format == 'channels_first':
            return -1 - self.rank
        else:
            return -1

    def _get_input_channel(self, input_shape):
        channel_axis = self._get_channel_axis()
        if input_shape.dims[channel_axis].value is None:
            raise ValueError('The channel dimension of the inputs '
                           'should be defined. Found `None`.')
        return int(input_shape[channel_axis])

    def _get_padding_op(self):
        if self.padding == 'causal':
            op_padding = 'valid'
        else:
            op_padding = self.padding
        if not isinstance(op_padding, (list, tuple)):
            op_padding = op_padding.upper()
        return op_padding





#Author: Joydeep Munshi
#Augmentation utility funciton for crystal4D
"""
This is augmentation pipeline for 4DSTEM/STEM diffraction (CBED) images. The different available augmentation includes elliptic distrotion, plasmonic background noise and poisson shot noise. The elliptic distortion is recommeded to be applied on CBED, Probe, Potential Vg and Qz tilts while other noise are only applied on CBED input.
"""
import time
import numpy as np
import tensorflow as tf
import scipy.signal as sp
import numpy.matlib as nm
from itertools import product
from ..utils import aug_utils
from .interpolate import dense_image_warp

from tensorflow.python.framework import tensor_shape

class image_augmentation(object):
    def __init__(self, 
                 backgrnd = False,
                 shot =False,
                 pattern_shift = False,
                 ellipticity = False,
                 counts_per_pixel = 1000,
                 qBackgroundLorentz = 0.1,
                 weightBackgrnd = 0.1,
                 scale = 1.0,
                 xshift = 1,
                 yshift = 1,
                 verbose = False,
                 log_file = './logs/augment_log.csv'):
        self.backgrnd = backgrnd
        self.shot = shot
        self.pattern_shift = pattern_shift
        self.ellipticity = ellipticity
        
        if self.backgrnd:
            print('background')
            self.weightBackgrnd = weightBackgrnd
            self.qBackgroundLorentz = qBackgroundLorentz
        if self.shot:
            self.counts_per_pixel = counts_per_pixel
        if self.pattern_shift:
            self.xshift = xshift
            self.yshift = yshift
        if self.ellipticity:
            self.scale = scale
        
        self.verbose = verbose
        self.log_file = log_file
        
        file_object = open(self.log_file, 'a')
        file_object.write('ellipticity,shot,background,pattern_shift \n')
        file_object.close()

    def set_params(self, 
                 backgrnd = True, 
                 shot =True, 
                 pattern_shift = True,
                 ellipticity = True,
                 counts_per_pixel = 1000,
                 qBackgroundLorentz = 0.1,
                 weightBackgrnd = 0.1,
                 scale = 1.0,
                 xshift = 1,
                 yshift = 1):
        
        self.backgrnd = backgrnd
        self.shot = shot
        self.pattern_shift = pattern_shift
        self.ellipticity = ellipticity
        
        if self.backgrnd:
            self.weightBackgrnd = weightBackgrnd
            self.qBackgroundLorentz = qBackgroundLorentz
        if self.shot:
            self.counts_per_pixel = counts_per_pixel
        if self.pattern_shift:
            self.xshift = xshift
            self.yshift = yshift
        if self.ellipticity:
            self.scale = scale
        
    def get_params(self): 
        print('Printing augmentation summary... \n',end = "\r")
        print('>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> \n',end = "\r")
        print('Shots per pixel: {} \n'.format(self.counts_per_pixel),end = "\r")
        print('Background plasmon: {} \n'.format(self.qBackgroundLorentz),end = "\r")
        print('Ellipticity scaling: {} \n'.format(self.scale),end = "\r")
        print('Pattern shift: {},{} \n'.format(self.xshift,self.yshift),end = "\r")
        print('<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< \n', end = "\r")
        
    
    def write_logs(self):
        file_object = open(self.log_file, 'a')
        file_object.write('{},{},{},{} \n'.format(self.ellipticity,self.shot,self.backgrnd,self.pattern_shift))
        file_object.close()
        

    def _get_dim(self, x, idx):
        if x.shape.ndims is None:
            return tf.shape(x)[idx]
        return x.shape[idx] or tf.shape(x)[idx]
    
    
    def _get_qx_qy(self, input_shape, pixel_size_AA = 0.20):
        """
        get qx,qy from cbed
        """
        N = input_shape
        qx = np.sort(np.fft.fftfreq(N[0], pixel_size_AA)).reshape((N[0], 1, 1))
        qy = np.sort(np.fft.fftfreq(N[1], pixel_size_AA)).reshape((1, N[1], 1))
        
        return qx, qy
    
    def _make_fourier_coord(self, Nx, Ny, pixelSize):
        """
        Generates Fourier coordinates for a (Nx,Ny)-shaped 2D array.
        Specifying the pixelSize argument sets a unit size.
        """
        if hasattr(pixelSize, '__len__'):
            assert len(pixelSize) == 2, "pixelSize must either be a scalar or have length 2"
            pixelSize_x = pixelSize[0]
            pixelSize_y = pixelSize[1]
        else:
            pixelSize_x = pixelSize
            pixelSize_y = pixelSize

        qx = np.fft.fftfreq(Nx, pixelSize_x)
        qy = np.fft.fftfreq(Ny, pixelSize_y)
        qy, qx = np.meshgrid(qy, qx)
        return qx, qy
    
    @tf.function
    def augment_img(self, inputs: aug_utils.TensorLike, probe = None):
        start_time = time.time()
        input_shape = inputs.shape
        if self.backgrnd:
            input_noise = self.backgrnd_plasmon(inputs, probe)
        else:
            input_noise = inputs
            self.weightBackgrnd = 0
            self.qBackgroundLorentz = 0
        
        if self.shot:
            input_noise = self.shot_noise(input_noise)
        else:
            self.counts_per_pixel = 0
               
        ################
        if self.pattern_shift:
            #Added on 09/22/2021 - Joydeep Munshi
            input_noise = self.pattern_shift_ar(input_noise)
        else:
            self.xhsift = 0
            self.yshift = 0
        
        if self.ellipticity:
            input_noise = self.elliptic_distort(input_noise)
        else:
            self.scale = 0
        
        t = time.time() - start_time
        t = t/60
        
        if self.verbose:
            self.get_params()
            self.write_logs()
            print('Augmentation Status: it took {} minutes to augment {} images... \n'.format(t, input_shape[0]), end = "\r")
        else:
            self.write_logs()
            
        return input_noise
    
    @tf.function
    def scale_image(self, inputs: aug_utils.TensorLike):
        '''
        Scale image between 0 and 1
        '''
        input_shape = tf.shape(inputs)
        mean = tf.squeeze(tf.math.reduce_sum(inputs, axis = (1,2)))
        inputs_scaled = tf.transpose(tf.transpose(inputs, [3,1,2,0])/mean, [3,1,2,0])
        
        return inputs_scaled
    
    @tf.function
    def shot_noise(self, inputs: aug_utils.TensorLike):
        """
        Apply Shot noise
        """
        image = tf.convert_to_tensor(inputs)
        #image_scale = self.scale_image(inputs)
        image_noise = tf.random.poisson(shape = [], lam = tf.maximum(image,0) * self.counts_per_pixel)/float(self.counts_per_pixel)
        
        return image_noise
    
    @tf.function
    def elliptic_distort(
        self, inputs: aug_utils.TensorLike):
        """
        Elliptic distortion
        """
        image = tf.convert_to_tensor(inputs)
        batch_size, height, width, channels = (
            self._get_dim(image, 0),
            self._get_dim(image, 1),
            self._get_dim(image, 2),
            self._get_dim(image, 3),
        )
        
        exx = 0.7 * self.scale
        eyy = -0.5 * self.scale
        exy = -0.9 * self.scale
        
        m = [[1+exx, exy/2], [exy/2, 1+eyy]]
        
        grid_x, grid_y = tf.meshgrid(tf.range(width), tf.range(height))
        grid_x = grid_x - tf.math.reduce_mean(grid_x)
        grid_y = grid_y - tf.math.reduce_mean(grid_y)
        
        grid_x = tf.cast(grid_x, tf.float32)
        grid_y = tf.cast(grid_y, tf.float32)
        
        #print(stacked_grid)

        flow_x =  (grid_x*m[0][0] + grid_y*m[1][0])
        flow_y =  (grid_x*m[0][1] + grid_y*m[1][1])
        
        flow_grid = tf.cast(tf.stack([flow_y, flow_x], axis=2), grid_x.dtype)
        batched_flow_grid = tf.expand_dims(flow_grid, axis=0)
        
        batched_flow_grid = tf.repeat(batched_flow_grid, batch_size, axis=0)
        #assert(self._get_dim(batched_flow_grid, 0) == batch_size); "The flow batch size and the image batch size is different!!"
        
        imageOut = dense_image_warp(image, batched_flow_grid)
       
        return imageOut
    
    @tf.function
    def backgrnd_plasmon(self, inputs: aug_utils.TensorLike, probe: aug_utils.TensorLike):
        """
        Apply background plasmon noise
        """
        image = tf.convert_to_tensor(inputs)
        batch_size, height, width, channels = (
            self._get_dim(image, 0),
            self._get_dim(image, 1),
            self._get_dim(image, 2),
            self._get_dim(image, 3),
        )
        
        qx,qy = self._get_qx_qy([height,width])
        CBEDbg = 1./ (qx**2 + qy**2 + self.qBackgroundLorentz**2)
        CBEDbg = tf.squeeze(CBEDbg)
        CBEDbg = CBEDbg / tf.math.reduce_sum(CBEDbg)

        CBEDbg = tf.expand_dims(CBEDbg, axis=0)
        CBEDbg = tf.repeat(CBEDbg, batch_size, axis=0)
        CBEDbg = tf.expand_dims(CBEDbg, axis=-1)
        CBEDbg = tf.repeat(CBEDbg, channels, axis=0)
        
        CBEDbg = tf.keras.backend.permute_dimensions(CBEDbg, (3,0,1,2))
        probe = tf.keras.backend.permute_dimensions(probe, (3,0,1,2))

        CBEDbg_ff = tf.cast(tf.signal.rfft2d(CBEDbg), tf.complex128)
        probe_ff = tf.cast(tf.signal.rfft2d(probe), tf.complex128)
        mul_ff = tf.multiply(CBEDbg_ff, probe_ff)
        
        CBEDbgConv = tf.signal.fftshift(tf.signal.irfft2d(mul_ff), axes = [2,3])
        CBEDbgConv = tf.keras.backend.permute_dimensions(CBEDbgConv, (1,2,3,0))
        
        CBEDout = tf.cast(inputs, tf.float32) * (1-self.weightBackgrnd) + tf.cast(CBEDbgConv, tf.float32) * self.weightBackgrnd
        
        return CBEDout
    
    @tf.function
    def pattern_shift_ar(self, inputs: aug_utils.TensorLike):
        """
        Apply pixel shift to the pattern using Fourier shift theorem for subpixel shifting
        """
        image = tf.convert_to_tensor(inputs)
        batch_size, height, width, channels = (
            self._get_dim(image, 0),
            self._get_dim(image, 1),
            self._get_dim(image, 2),
            self._get_dim(image, 3),
        )
        
        qx, qy = self._make_fourier_coord(height, width, 1)
        
        ar = tf.cast(tf.keras.backend.permute_dimensions(image, (0,3,1,2)), tf.complex64)

        w = np.exp(-(2j * np.pi) * ((self.yshift * qy) + (self.xshift * qx)))
        shifted_ar = tf.math.real(tf.signal.ifft2d(tf.math.multiply(tf.signal.fft2d(ar), w)))
        
        shifted_ar = tf.keras.backend.permute_dimensions(shifted_ar, (0,2,3,1))
        
        return shifted_ar